#include <Arduino.h>
#include <LowPower.h>
#include <EEPROM.h>
#include <AES.h>
#include <AES_config.h>
#include <nRF24L01.h>
#include <printf.h>
#include <RF24.h>
#include <RF24_config.h>
#include <stdint.h>
#include <WeatherStationPacket.h>

/* -------CONFIG---------*/
#define LOGDEBUG        0
#define SLEEP_LOOP_MAX  300
#define VOLTAGE_RATIO   4.9746f  //(1100/1023)*((33+9.1)/9.1) 
/* --------------------- */


/* -------LOGDEBUG------ */
#if LOGDEBUG==1
#define LOGINFO(...) Serial.printf(__VA_ARGS__)
#else
#define LOGINFO(...)
#endif
/* --------------------- */


/* -------PINS---------- */
#define SUPERCAP_ADC_ENABLE   16
#define SUPERCAP_ADC          A1 //15
#define LED                   14
#define RF24_CSN              5
#define RF24_CE               6
#define PULSE0                2
#define PULSE1                3
/* --------------------- */


/* -------WIRELESS------ */
#define RF24_WEATHER_CH       73
/* --------------------- */


/* ---WIRELESS ADDRESS--- */
uint8_t   weatherGatewayAddress[5]  = {0, 0, 0, 0, 0}; //Set your NRF24 receiver address here
uint8_t   weatherSensorAddress[5]   = {1, 0, 0, 0, 0}; //Set your NRF24 transmiter address here (this device)
/* --------------------- */


/* -------AES KEYS------ */
uint8_t   weatherAesKey[32] = {0x6d, 0x56, 0x1c, 0x2d, 0xf0, 0xe4, 0xc5, 0x97, 0x81, 0x8a, 0x27, 0x88, 0x5d, 0xd3, 0x8d, 0x10, 0xf0, 0x37, 0x8e, 0x56, 0x51, 0x02, 0x14, 0x80, 0xc3, 0xb8, 0x78, 0x3d, 0x2f, 0x9f, 0x09 , 0x11}; 
/* --------------------- */
                               

WeatherStationPacket weatherStationPacket(2);
AES       aes;
RF24      radio(RF24_CE, RF24_CSN, 2000000);

uint16_t sleepLoop = SLEEP_LOOP_MAX;

uint32_t rainCounter = 0;
bool updateEepromNeeded = false;
uint8_t rainCounterAddress = 0;


uint16_t getCapacitorVoltage(){
  //enable voltage divisor
  LOGINFO("enable voltage divisor\n");
  digitalWrite(SUPERCAP_ADC_ENABLE, HIGH);
  delay(10);

  float voltage = analogRead(SUPERCAP_ADC);
  voltage = voltage * VOLTAGE_RATIO; 

  //disable voltage divisor
  LOGINFO("disable voltage divisor\n");
  digitalWrite(SUPERCAP_ADC_ENABLE, LOW);  

  LOGINFO("Voltage value: %hu\n", (uint16_t)voltage);
  return (uint16_t)voltage;
}

void updateRainCounter(){
	++rainCounter;
  LOGINFO("rainCounter=%lu\n", rainCounter);
	updateEepromNeeded = true;
}

void resetEEPROM(){  
  for (uint32_t i = 0 ; i < EEPROM.length() ; i++) {
    EEPROM.write(i, 0);
    LOGINFO("write eeprom at %lu\n", i);
  }
}

void writeRainCounterToEEPROM(){
	//circular buffer using all eeprom space (1024 bytes -> 256 x uint32_t)
	LOGINFO("setRainCounterToEEPROM\n");
	if(rainCounterAddress == 0xFF){
		rainCounterAddress = 0;
	} else {
		++rainCounterAddress;
	}
  LOGINFO("Write value %lu at address %u\n", rainCounter, rainCounterAddress);
	EEPROM.put(4l*rainCounterAddress, rainCounter);
}

void readRainCounterFromEEPROM(){
  LOGINFO("getRainCounterFromEEPROM\n");
  uint32_t tmpRainCounter = 0;
  for(uint8_t i = 0; i < 0xFF; ++i){
    EEPROM.get(4l*i, tmpRainCounter);
    if(tmpRainCounter > rainCounter){
      rainCounter = tmpRainCounter;
      rainCounterAddress = i;
      LOGINFO("Found value %lu at address %u\n", rainCounter, rainCounterAddress);
    }
  }
  LOGINFO("Final value found %lu at address %u\n", rainCounter, rainCounterAddress);
}


void setup() {
  delay(500);  

  pinMode(SUPERCAP_ADC_ENABLE , OUTPUT);
  digitalWrite(SUPERCAP_ADC_ENABLE, LOW);

  pinMode(LED , OUTPUT);
  digitalWrite(LED, HIGH);

  analogReference(INTERNAL1V1);

  #if LOGDEBUG==1
  Serial.begin(250000);  
  printf_begin();
  #endif

  LOGINFO("Starting...\n");

  //resetEEPROM();

  radio.begin();
  radio.setChannel(RF24_WEATHER_CH);
  radio.setDataRate(RF24_250KBPS);
  radio.setPALevel(RF24_PA_MAX);
  radio.setCRCLength(RF24_CRC_16);
  radio.setRetries(15, 5);  
  radio.setAutoAck(1);
  radio.openReadingPipe(1, weatherSensorAddress);
  radio.openWritingPipe(weatherGatewayAddress);

  #if LOGDEBUG==1
  radio.printDetails();
  #endif

  aes.set_key(weatherAesKey, 256);
  
  readRainCounterFromEEPROM();
  
  pinMode(PULSE0, INPUT);
  attachInterrupt(digitalPinToInterrupt(PULSE1), updateRainCounter, FALLING);

  delay(500);
  digitalWrite(LED, LOW);
}



void loop() {  
  if( sleepLoop >= SLEEP_LOOP_MAX){
	  
	if(updateEepromNeeded){
		cli();
    LOGINFO("Update rain counter to: %lu\n", rainCounter);
		writeRainCounterToEEPROM();
		updateEepromNeeded = false;
		sei();
	}
	
    uint16_t supercapVoltage = getCapacitorVoltage();

    //Send to weather station
    LOGINFO("Send to weather station\n");
    weatherStationPacket.setType(0);
    weatherStationPacket.setStatus(0);

    weatherStationPacket.write(4, &rainCounter, 4);
    weatherStationPacket.write(8, &supercapVoltage, 2);

    weatherStationPacket.updateCRC();


  #if LOGDEBUG==1
    weatherStationPacket.printBuffer();
  #endif
    
    aes.encrypt(weatherStationPacket.getPlainBuffer(),     weatherStationPacket.getCipherBuffer());
    aes.encrypt(weatherStationPacket.getPlainBuffer()+16,     weatherStationPacket.getCipherBuffer()+16);

    radio.setChannel(RF24_WEATHER_CH);
    radio.openReadingPipe(1, weatherSensorAddress);
    radio.openWritingPipe(weatherGatewayAddress);
    
    radio.powerUp();
    delay(10); 
    radio.write(weatherStationPacket.getCipherBuffer(), 32);
    radio.powerDown();
   
    
    sleepLoop = 0;
  } else {
    #if LOGDEBUG==1
    LOGINFO(".");
    delay(10);  
    #endif    
  
    ++sleepLoop;
  }

  LowPower.powerDown(SLEEP_1S, ADC_OFF, BOD_ON);
}
